unit regmashine;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, jkTools, contnrs;

type
  TOpcode=(inSTO, inLDA, inBRZ, inADD, inSUB, inOR, inAND, inXOR, inNOT, inINC, inDEC, inZRO, inNOP);
  tOperation=class
    private
      FOpcode: TOpcode;
      Fadress: word;
      function toString:string;
      function getAssembled:word;
    public
      constructor create; overload;
      constructor create(instr:string); overload;
      constructor create(opc:TOpcode; adr:word); overload;
    published
      property Opcode:TOpcode read FOpcode;
      property Adress:word read Fadress;
      property asString:string read toString;
      property Assembled:word read getAssembled;
  end;


  TRegMashine = class
  private
    Fmemory_separation:word;
    Fmembox1, Fmembox2:TListBox;
    FMemory:Array[0..$FFF+10] of word;
    FCurrent, FmaxMem:word;
    FAkku:word;
    procedure mem1DrawItem(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState);
    procedure mem2DrawItem(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState);
    function getCurrent:string;
    procedure setSep(value:word);
    procedure UpdateView(pos:word);
  public
    constructor create(membox1, membox2:TListBox);
    destructor destroy;
    procedure Parse(prog:TStrings);
    procedure Step;
    procedure InitializeView;

  published
    property memory_separation:word read Fmemory_separation write SetSep;
    property CurrentInstruction:word read FCurrent;
    property CurrentInstructionStr:string read getCurrent;
    property Akku:word read FAkku;
end;

function containsInstructions(instr:string):boolean;
function intToInstructionStr(mem:word):string;
function intToInstruction(mem:word):tOperation;

implementation

function intToInstruction(mem:word):tOperation;
var op:TOperation;
    opcode:TOpcode;
    adress, instr:word;
begin
  adress:=mem and $0FFF;
  instr:=mem and $F000;
  instr:=instr shr 12;
  opcode:=inNOP;
  case instr of
    0: opcode:=inSTO;
    1: opcode:=inLDA;
    2: opcode:=inBRZ;
    3: opcode:=inADD;
    4: opcode:=inSUB;
    5: opcode:=inOR;
    6: opcode:=inAND;
    7: opcode:=inXOR;
    8: opcode:=inNOT;
    9: opcode:=inINC;
    10: opcode:=inDEC;
    11: opcode:=inZRO;
    12..15: opcode:=inNOP;
  end;
  op:=TOperation.create(opcode, adress);
  result:=op;
end;

function intToInstructionStr(mem:word):string;
var op:TOperation;
begin
  op:=intToInstruction(mem);
  result:=op.asString;
  op.free;
end;


function containsInstructions(instr:string):boolean;
var data, ins, adr:string;
    i:word;
    isInstruction:boolean;
    isComment:boolean;
begin
  result:=false;
  isComment:=false;
  data:='';

  // strip comments from instructions
  for i:=1 to length(instr) do begin
    if instr[i] in [';', '#'] then isComment:=true;
    if not isComment then data:=data+ansilowercase(instr[i]);
  end;

  // parse assembler-line into instruction and adress.
  ins:='';
  adr:='';
  isInstruction:=true;
  if length(data)>0 then for i:=1 to length(data) do begin
    if (data[i] in [#32, #9])and(ins<>'') then isInstruction:=false;
    if isInstruction and (data[i] in ['a'..'z']) then result:=true;
  end;

end;

constructor tOperation.create;
begin
  FOpcode:=inNOP;
  FAdress:=0;
end;

constructor tOperation.create(instr:string);
var data, ins, adr:string;
    i:word;
    isInstruction:boolean;
    isComment:boolean;
begin
  isComment:=false;
  data:='';

  // strip comments from instructions
  for i:=1 to length(instr) do begin
    if instr[i] in [';', '#'] then isComment:=true;
    if not isComment then data:=data+ansilowercase(instr[i]);
  end;

  // parse assembler-line into instruction and adress.
  ins:='';
  adr:='';
  isInstruction:=true;
  if length(data)>0 then for i:=1 to length(data) do begin
    if (data[i] in [#32, #9])and(ins<>'') then isInstruction:=false;
    if isInstruction and (data[i] in ['a'..'z']) then ins:=ins+data[i];
    if (not isInstruction) and (data[i] in ['a'..'f', '0'..'9', 'x', '$']) then adr:=adr+data[i];
  end;

  //trying to recognize instructions
  FOpcode:=inNOP;
  FAdress:=0;
  if ins='sto' then FOpcode:=inSTO;
  if ins='lda' then FOpcode:=inLDA;
  if ins='brz' then FOpcode:=inBRZ;
  if ins='add' then FOpcode:=inADD;
  if ins='sub' then FOpcode:=inSUB;
  if ins='or' then FOpcode:=inOR;
  if ins='and' then FOpcode:=inAND;
  if ins='xor' then FOpcode:=inXOR;
  if ins='not' then FOpcode:=inNOT;
  if ins='inc' then FOpcode:=inINC;
  if ins='dec' then FOpcode:=inDEC;
  if ins='zro' then FOpcode:=inZRO;
  if ins='nop' then FOpcode:=inNOP;

  if FOpcode in [inSTO, inLDA, inBRZ, inADD, inSUB, inOR, inAND, inXOR] then try
    FAdress:=strtoint(adr);
  except
    ShowErrorMessage(instr+#13#10'== '+data+#13#10'may not be parsed!')
  end;

end;

constructor tOperation.create(opc:TOpcode; adr:word);
begin
  FOpcode:=opc;
  FAdress:=adr;
end;

function tOperation.toString:string;
begin
  case FOpcode of
    inSTO: result:='STO ';
    inLDA: result:='LDA ';
    inBRZ: result:='BRZ ';
    inADD: result:='ADD ';
    inSUB: result:='SUB ';
    inOR: result:='OR  ';
    inAND: result:='AND ';
    inXOR: result:='XOR ';
    inNOT: result:='NOT ';
    inINC: result:='INC ';
    inDEC: result:='DEC ';
    inZRO: result:='ZRO ';
    inNOP: result:='NOP ';
  end;
  if FOpcode in [inSTO, inLDA, inBRZ, inADD, inSUB, inOR, inAND, inXOR] then result:=result+'0x'+inttohex(Fadress,3);
end;

function tOperation.getAssembled:word;
begin
  case FOpcode of
    inSTO: result:=0;
    inLDA: result:=1;
    inBRZ: result:=2;
    inADD: result:=3;
    inSUB: result:=4;
    inOR: result:=5;
    inAND: result:=6;
    inXOR: result:=7;
    inNOT: result:=8;
    inINC: result:=9;
    inDEC: result:=10;
    inZRO: result:=11;
    inNOP: result:=12;
  end;
  result:=result shl 12;
  if FOpcode in [inSTO, inLDA, inBRZ, inADD, inSUB, inOR, inAND, inXOR] then result:=result+Fadress;
end;





constructor TRegMashine.create(membox1, membox2:TListBox);
var i:word;
begin
  Fmemory_separation:=$40;
  FmaxMem:=Fmemory_separation;
  Fmembox1:=membox1;
  Fmembox2:=membox2;
  FCurrent:=0;
  FAkku:=0;
  for i:=0 to $FFF do
    FMemory[i]:=0;
end;

destructor TRegMashine.destroy;
begin
end;

procedure TRegMashine.Parse(prog:TStrings);
var i:word;
    op:TOperation;
    adr:word;
begin
  adr:=0;
  for i:=0 to $FFF do
    FMemory[i]:=0;
  if prog.count>0 then for i:=0 to prog.count do begin
    if containsInstructions(prog[i]) then begin
      op:=TOperation.create(prog[i]);
      FMemory[adr]:=op.Assembled;
      adr:=adr+1;
      op.free;
    end;
  end;
  FCurrent:=0;
  FAkku:=0;
end;

procedure TRegMashine.Step;
var op:TOperation;
    old:word;
begin
  op:=intToInstruction(FMemory[FCurrent]);
  old:=FCurrent;
  case op.Opcode of
    inSTO:  begin
              FMemory[op.Adress]:=FAkku;
              FCurrent:=FCurrent+1;
              UpdateView(op.Adress)
            end;
    inLDA:  begin
              FAkku:=FMemory[op.Adress];
              FCurrent:=FCurrent+1;
            end;
    inBRZ:  begin
              FCurrent:=FCurrent+1;
              If FAkku=0 then FCurrent:=op.Adress;
            end;
    inADD:  begin
              FAkku:=FAkku+FMemory[op.Adress];
              FCurrent:=FCurrent+1;
            end;
    inSUB:  begin
              FAkku:=FAkku-FMemory[op.Adress];
              FCurrent:=FCurrent+1;
            end;
    inOR :  begin
              FAkku:=FAkku or FMemory[op.Adress];
              FCurrent:=FCurrent+1;
            end;
    inAND:  begin
              FAkku:=FAkku and FMemory[op.Adress];
              FCurrent:=FCurrent+1;
            end;
    inXOR:  begin
              FAkku:=FAkku xor FMemory[op.Adress];
              FCurrent:=FCurrent+1;
            end;
    inINC:  begin
              FAkku:=FAkku+1;
              FCurrent:=FCurrent+1;
            end;
    inDEC:  begin
              FAkku:=FAkku-1;
              FCurrent:=FCurrent+1;
            end;
    inZRO:  begin
              FAkku:=0;
              FCurrent:=FCurrent+1;
            end;
    inNOP:  begin
              FCurrent:=FCurrent+1;
            end;
  end;
  UpdateView(FCurrent);
  UpdateView(old);
end;

procedure TRegMashine.InitializeView;
var i:word;
begin
  Fmembox1.Items.Clear;
  Fmembox2.Items.Clear;
  Fmembox1.MultiSelect:=false;
  Fmembox2.MultiSelect:=false;
  Fmembox1.OnDrawItem:=mem1DrawItem;
  Fmembox2.OnDrawItem:=mem2DrawItem;
  for i:=0 to Fmemory_separation do begin
    Fmembox1.Items.add('0x'+inttohex(i, 3)+':     0x'+inttohex(FMemory[i], 4)+' = '+intToInstructionStr(FMemory[i]));
  end;
  for i:=Fmemory_separation+1 to $FFF do begin
    Fmembox2.Items.add('0x'+inttohex(i, 3)+':     0x'+inttohex(FMemory[i], 4)+' = '+inttostr(FMemory[i]));
  end;
  //UpdateView;
end;

procedure TRegMashine.UpdateView(pos:word);
begin
  if pos>-1 then begin
    if pos<=Fmemory_separation then
      Fmembox1.Items[pos]:='0x'+inttohex(pos, 3)+':     0x'+inttohex(FMemory[pos], 4)+' = '+intToInstructionStr(FMemory[pos])
    else
      Fmembox2.Items[pos-Fmemory_separation-1]:='0x'+inttohex(pos, 3)+':     0x'+inttohex(FMemory[pos], 4)+' = '+inttostr(FMemory[pos]);
  end;
end;

procedure TRegMashine.mem1DrawItem(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState);
begin
  with (Control as TListBox).Canvas do begin
    Brush.color:=clWhite;
    Font.color:=clBlack;
    if (FCurrent=Index) then begin
      Brush.color:=clNavy;
      Font.color:=clWhite;
    end;
    FillRect(Rect);       { Rechteck lschen }
    TextOut(Rect.Left+2, Rect.Top, (Control as TListBox).Items[Index])  { Text anzeigen }
  end;
end;

procedure TRegMashine.mem2DrawItem(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState);
begin
  with (Control as TListBox).Canvas do begin
    Brush.color:=clWhite;
    Font.color:=clBlack;
    if (FCurrent=Index-1-FMemory_Separation) then begin
      Brush.color:=clNavy;
      Font.color:=clWhite;
    end;
    FillRect(Rect);       { Rechteck lschen }
    TextOut(Rect.Left+2, Rect.Top, (Control as TListBox).Items[Index])  { Text anzeigen }
  end;
end;

function  TRegMashine.getCurrent:string;
var op:TOperation;
begin
  op:=intToInstruction(FMemory[FCurrent]);
  result:=op.asString;
  op.free;
end;

procedure TRegMashine.setSep(value:word);
begin
  FMemory_Separation:=value;
end;

end.
